@@ -3,6 +3,13 @@
 
 """publish files on the internet"""
 
+__epilog__ = """This program will upload the provided files on the
+internet and notify the user when completed. The resulting URL will
+also be copied to the clipboard. Preprocessors can do fancy things
+before (only supports building an image gallery with sigal for
+now). Options can be specified in the config file without the leading
+dashes as well."""
+
 # This is a (partial) rewrite of weasel's "publish" program. See the
 # `parse_args` function for details on the behavior changes and
 # variations.
@@ -22,9 +29,6 @@
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import division, absolute_import
-from __future__ import print_function, unicode_literals
-
 import argparse
 import atexit
 from dataclasses import dataclass, astuple
@@ -34,10 +38,12 @@ import os.path
 from pathlib import Path
 import secrets  # python 3.6
 import shutil
+from site import USER_BASE
 import subprocess
 import sys
 import tempfile
 from typing import (
+    Any,
     Callable,
     Iterable,
     Iterator,
@@ -47,6 +53,7 @@ from typing import (
     Tuple,
     Type,
     TypeVar,
+    Union,
 )
 
 try:
@@ -91,16 +98,10 @@ except (ImportError, ValueError):
         pass
 
 
-__epilog__ = """ This program will upload the provided files on the internet and
-notify the user when completed. The resulting URL will also be copied
-to the clipboard. Preprocessors can do fancy things before (only
-supports building an image gallery with sigal for now)."""
-
-
 class LoggingAction(argparse.Action):
     """change log level on the fly
 
-    The logging system should be initialized befure this, using
+    The logging system should be initialized before this, using
     `basicConfig`.
     """
 
@@ -116,11 +117,20 @@ class LoggingAction(argparse.Action):
             kwargs["nargs"] = 0
         super().__init__(*args, **kwargs)
 
-    def __call__(self, parser, ns, values, option):  # type: ignore[no-untyped-def]
+    def __call__(
+        self,
+        parser: argparse.ArgumentParser,
+        ns: argparse.Namespace,
+        values: Optional[Union[str, Sequence[Any]]],
+        option_string: Optional[str] = None,
+    ) -> None:
         """if const was specified it means argument-less parameters"""
         if self.const:
             logging.getLogger("").setLevel(self.const)
         else:
+            values = str(values)
+            if values not in logging.getLevelNamesMapping().keys():
+                parser.error("invalid logging level: %s" % values)
             logging.getLogger("").setLevel(values)
         # cargo-culted from _StoreConstAction
         setattr(ns, self.dest, self.const or values)
@@ -131,25 +141,58 @@ class ConfigAction(argparse._StoreAction):
 
     a *list* of default config files can be specified and will be
     parsed when added by ConfigArgumentParser.
-
-    it was reported this might not work well with subparsers, patches
-    to fix that are welcome.
     """
 
     def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         """the config action is a search path, so a list, so one or more argument"""
         kwargs["nargs"] = 1
+        self._config_action_called = False
         super().__init__(*args, **kwargs)
 
-    def __call__(self, parser, ns, values, option):  # type: ignore[no-untyped-def]
+    def __call__(
+        self,
+        parser: argparse.ArgumentParser,
+        ns: argparse.Namespace,
+        values: Optional[Union[str, Sequence[Any]]] = None,
+        option_string: Optional[str] = None,
+    ) -> None:
         """change defaults for the namespace, still allows overriding
         from commandline options"""
-        for path in values:
-            parser.set_defaults(**self.parse_config(path))
-        super().__call__(parser, ns, values, option)
+        if self._config_action_called:
+            return
+        if values:
+            if not isinstance(values, Sequence):
+                values = [values]
+            saved_exception: Exception | None = None
+            for path in values:
+                try:
+                    parser.set_defaults(**self.parse_config(path))
+                    self._config_action_called = True
+                except OSError as e:
+                    logging.debug("failed to load config file %s: %s", path, e)
+                    saved_exception = e
+                except argparse.ArgumentError as e:
+                    logging.warning("failed to parse config file %s: %s", path, e)
+                    saved_exception = e
+                else:
+                    # stop processing once we find a valid configuration
+                    # file
+                    break
+            else:
+                # re-raise failure if we really did not find any
+                # configuration file
+                if saved_exception:
+                    raise saved_exception
+        super().__call__(parser, ns, values, option_string)
 
     def parse_config(self, path: str) -> dict:  # type: ignore[type-arg]
-        """abstract implementation of config file parsing, should be overriden in subclasses"""
+        """abstract implementation of config file parsing, should be overridden in subclasses
+
+        Parse errors should raise argparse.ArgumentError to report
+        parse errors as a usage error to the caller and leave OSError
+        to be raised, as they are expected and handled by the
+        ConfigArgumentParser.
+        """
         raise NotImplementedError()
 
 
@@ -157,12 +200,15 @@ class YamlConfigAction(ConfigAction):
     """YAML config file parser action"""
 
     def parse_config(self, path: str) -> dict:  # type: ignore[type-arg]
+        """Raises YAML parse errors as argument errors."""
         try:
             with open(os.path.expanduser(path), "r") as handle:
                 logging.debug("parsing path %s as YAML" % path)
                 return yaml.safe_load(handle) or {}
-        except (FileNotFoundError, yaml.parser.ParserError) as e:
-            raise argparse.ArgumentError(self, str(e))
+        except yaml.error.YAMLError as e:
+            raise argparse.ArgumentError(
+                self, "failed to parse YAML configuration: %s" % str(e)
+            )
 
 
 class ConfigArgumentParser(argparse.ArgumentParser):
@@ -177,6 +223,9 @@ class ConfigArgumentParser(argparse.ArgumentParser):
 
             self.add_argument('--config', action=YamlConfigAction, default=self.default_config())
 
+    This shows how the configuration file overrides the default value
+    for an option:
+
     >>> from tempfile import NamedTemporaryFile
     >>> c = NamedTemporaryFile()
     >>> c.write(b"foo: delayed\\n")
@@ -190,10 +239,12 @@ class ConfigArgumentParser(argparse.ArgumentParser):
     True
     >>> args.foo
     'delayed'
+    >>> args = parser.parse_args(['--foo', 'quux'])
+    >>> args.foo
+    'quux'
 
     This is the same test, but with `--config` called earlier, which
-    should still work (but doesn't, because the subsequent action
-    changes the default).
+    should still work:
 
     >>> from tempfile import NamedTemporaryFile
     >>> c = NamedTemporaryFile()
@@ -211,6 +262,9 @@ class ConfigArgumentParser(argparse.ArgumentParser):
     >>> args = parser.parse_args(['--foo', 'baz'])
     >>> args.foo
     'baz'
+
+    This tests that you can override the config file defaults altogether:
+
     >>> parser = ConfigArgumentParser()
     >>> a = parser.add_argument('--config', action=YamlConfigAction, default=[c.name])
     >>> a = parser.add_argument('--foo', default='bar')
@@ -220,33 +274,35 @@ class ConfigArgumentParser(argparse.ArgumentParser):
     >>> args = parser.parse_args(['--config', '/dev/null', '--foo', 'baz'])
     >>> args.foo
     'baz'
+
+    This tests multiple search paths, first one should be loaded:
+
+    >>> from tempfile import NamedTemporaryFile
+    >>> d = NamedTemporaryFile()
+    >>> d.write(b"foo: argh\\n")
+    10
+    >>> d.flush()
+    >>> parser = ConfigArgumentParser()
+    >>> a = parser.add_argument('--config', action=YamlConfigAction, default=[d.name, c.name])
+    >>> a = parser.add_argument('--foo', default='bar')
+    >>> args = parser.parse_args([])
+    >>> args.foo
+    'argh'
     >>> c.close()
+    >>> d.close()
 
+    There are actually many other implementations of this we might
+    want to consider instead of maintaining our own:
+
+    https://github.com/omni-us/jsonargparse
+    https://github.com/bw2/ConfigArgParse
+    https://github.com/omry/omegaconf
+
+    See this comment for a quick review:
+
+    https://github.com/borgbackup/borg/issues/6551#issuecomment-1094104453
     """
 
-    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
-        super().__init__(*args, **kwargs)
-        # a list of actions to fire with their defaults if not fired
-        # during parsing
-        self._delayed_config_action = []
-
-    def _add_action(self, action: argparse.Action) -> argparse.Action:  # type: ignore[override]
-        # this overrides the add_argument() routine, which is where
-        # actions get registered in the argparse module.
-        #
-        # we do this so we can properly load the default config file
-        # before the the other arguments get set.
-        #
-        # now, of course, we do not fire the action here directly
-        # because that would make it impossible to *not* load the
-        # default action. so instead we register this as a
-        # "_delayed_config_action" which gets fired in `parse_args()`
-        # instead
-        action = super()._add_action(action)
-        if isinstance(action, ConfigAction) and action.default is not None:
-            self._delayed_config_action.append(action)
-        return action
-
     def parse_args(  # type: ignore[override]
         self,
         args: Optional[Sequence[str]] = None,
@@ -255,7 +311,10 @@ class ConfigArgumentParser(argparse.ArgumentParser):
         # we do a first failsafe pass on the commandline to find out
         # if we have any "config" parameters specified, in which case
         # we must *not* load the default config file
-        ns, _ = self.parse_known_args(args, namespace)
+        try:
+            ns, _ = self.parse_known_args(args, namespace)
+        except OSError as e:
+            self.error("failed to parse configuration file: %s" % e)
 
         # load the default configuration file, if relevant
         #
@@ -266,30 +325,42 @@ class ConfigArgumentParser(argparse.ArgumentParser):
         # we do this instead of just loading the config file in the
         # namespace precisely to make it possible to override the
         # configuration file settings on the commandline
-        for action in self._delayed_config_action:
+        for action in self._actions:
+            if not isinstance(action, ConfigAction) or action.default is None:
+                continue
             if action.dest in ns and action.default != getattr(ns, action.dest):
                 # do not load config default if specified on the commandline
-                logging.debug("not loading delayed action because of config override")
+                logging.debug("not loading default config because of --config override")
                 # action is already loaded, no need to parse it again
                 continue
             try:
                 action(self, ns, action.default, None)
                 logging.debug("loaded config file: %s" % action.default)
-            except argparse.ArgumentError as e:
-                # ignore errors from missing default
+            except OSError as e:
+                # ignore errors from missing default config file
                 logging.debug("default config file %s error: %s" % (action.default, e))
         # this will actually load the relevant config file when found
         # on the commandline
         #
         # note that this will load the config file a second time
-        return super().parse_args(args, namespace)
+        try:
+            return super().parse_args(args, namespace)
+        except OSError as e:
+            self.error("failed to parse configuration file: %s" % e)
 
     def default_config(self) -> Iterable[str]:
-        """handy shortcut to detect commonly used config paths"""
+        """handy shortcut to detect commonly used config paths
+
+        This list is processed as a FIFO: if a file is found in there,
+        it will be parsed and the remaining ones will be ignored.
+        """
         return [
             os.path.join(
                 os.environ.get("XDG_CONFIG_HOME", "~/.config/"), self.prog + ".yml"
-            )
+            ),
+            os.path.join(USER_BASE or "/usr/local", "etc", self.prog + ".yml"),
+            os.path.join("/usr/local/etc", self.prog + ".yml"),
+            os.path.join("/etc", self.prog + ".yml"),
         ]
 
 
@@ -298,7 +369,7 @@ class PubpasteArgumentParser(ConfigArgumentParser):
         """
         override constructor to setup our arguments and config file
         """
-        super().__init__(description=__doc__, epilog=__epilog__, *args, **kwargs)  # type: ignore
+        super().__init__(description=__doc__, epilog=__epilog__, add_help=False, *args, **kwargs)  # type: ignore
         # This is the usage of weasel's publish:
         #
         # usage: publish [<src> [<src> ...]]
@@ -323,84 +394,10 @@ class PubpasteArgumentParser(ConfigArgumentParser):
         #    -x        Publish the contents of the xclipboard.
         #    -u        Unpublish directory (only useful together with -T)
         #    -L        Follow symlinks
-        self.add_argument(
-            "-v",
-            "--verbose",
-            action=LoggingAction,
-            const="INFO",
-            help="enable verbose messages",
-        )
-        self.add_argument(
-            "-d",
-            "--debug",
-            action=LoggingAction,
-            const="DEBUG",
-            help="enable debugging messages",
-        )
-        self.add_argument("--dryrun", "-n", action="store_true", help="do nothing")
-
-        group = self.add_argument_group("history arguments")
-        group.add_argument(
-            "-t",
-            "--ttl",
-            help="how long to keep the entry, in days, default: %(default)s",
-        )
-        group = group.add_mutually_exclusive_group()
-        group.add_argument(
-            "-T", "--token", type=str, help="secret token, default: generated"
-        )
-        group.add_argument(
-            "-R", "--republish", action="store_true", help="reuse previous secret token"
-        )
-
-        self.add_argument(
-            "--follow-symlinks", "-L", action="store_true", help="follow symlinks"
-        )
-        self.add_argument(
-            "-r",
-            "--relative",
-            action="store_true",
-            help="consider all parts of the provided path, passed to rsync as --relative",
-        )
-        self.add_argument(
-            "-s",
-            "--stdin-name",
-            metavar="NAME",
-            default="stdin.txt",
-            type=str,
-            help="use NAME as filename when reading from stdin, default: %(default)s",
-        )
-        self.add_argument(
-            "-o",
-            "--output",
-            type=str,
-            help=argparse.SUPPRESS,  # only in configuration file
-            default=None,
-        )
-        self.add_argument(
-            "--url-prefix",
-            type=str,
-            help=argparse.SUPPRESS,  # only in configuration file
-            default=None,
-        )
-        self.add_argument(
-            "--base-dir",
-            type=str,
-            help=argparse.SUPPRESS,  # only in configuration file
-            default=None,
-        )
-        self.add_argument(
-            "--save-screenshots",
-            type=str,
-            help=argparse.SUPPRESS,  # only in configuration file
-        )
-        self.add_argument(
-            "--select",
-            action="store_true",
-            help="tell screenshot tool to let the user select an area, default: fullscreen",
-        )
+        self.add_argument("files", nargs="*", help="files to upload in default mode")
         commands_group = self.add_argument_group(
-            "command arguments", description="what should be done, default: just upload"
+            "commands",
+            description="what should be done, default: upload the given files",
         )
         group = commands_group.add_mutually_exclusive_group()
         group.add_argument(
@@ -469,17 +466,106 @@ class PubpasteArgumentParser(ConfigArgumentParser):
         )
         group.add_argument("--command", default="upload", help=argparse.SUPPRESS)
 
-        self.add_argument(
+        history_group = self.add_argument_group("history options")
+        history_group.add_argument(
+            "-t",
+            "--ttl",
+            help="how long to keep the entry, in days, default: %(default)s",
+        )
+        group = history_group.add_mutually_exclusive_group()
+        group.add_argument(
+            "-T", "--token", type=str, help="secret token, default: generated"
+        )
+        group.add_argument(
+            "-R", "--republish", action="store_true", help="reuse previous secret token"
+        )
+
+        config = self.add_argument_group("configuration options")
+        config.add_argument(
+            "-o",
+            "--output",
+            type=str,
+            help="rsync-compatible URL where to upload files, example: example.com:public_html/",
+            default=None,
+        )
+        config.add_argument(
+            "--url-prefix",
+            type=str,
+            help="how that output URL translates to a publicly visible URL, example: https://example.com/~user/",
+            default=None,
+        )
+        config.add_argument(
+            "--base-dir",
+            type=str,
+            help="local directory the output corresponds to, used in purge, example: /home/user/public_html/",
+            default=None,
+        )
+        config.add_argument(
+            "--save-screenshots",
+            type=str,
+            help="save a copy of screenshots in the given directory",
+        )
+        config.add_argument(
+            "--select",
+            action="store_true",
+            help="tell screenshot tool to let the user select an area, default: fullscreen",
+        )
+
+        options = self.add_argument_group("other options")
+        options.add_argument(
+            "-h",
+            "--help",
+            action="help",
+            default=argparse.SUPPRESS,
+            help="show this help message and exit",
+        )
+        options.add_argument(
+            "-v",
+            "--verbose",
+            action=LoggingAction,
+            const="INFO",
+            help="enable verbose messages",
+        )
+        options.add_argument(
+            "-d",
+            "--debug",
+            action=LoggingAction,
+            const="DEBUG",
+            help="enable debugging messages",
+        )
+        options.add_argument("--dryrun", "-n", action="store_true", help="do nothing")
+        options.add_argument(
+            "--follow-symlinks",
+            "-L",
+            action="store_true",
+            help="follow symlinks when copying files around",
+        )
+        options.add_argument(
+            "-r",
+            "--relative",
+            action="store_true",
+            help="consider all parts of the provided path, passed to rsync as --relative",
+        )
+        options.add_argument(
+            "-s",
+            "--stdin-name",
+            metavar="NAME",
+            default="stdin.txt",
+            type=str,
+            help="use NAME as filename when reading from stdin, default: %(default)s",
+        )
+        options.add_argument(
             "--config",
             action=YamlConfigAction,
             default=self.default_config(),
             help="use alternatte config file path, default: %(default)s",
         )
-        self.add_argument("files", nargs="*", help="files to upload")
 
     def parse_args(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         res = super().parse_args(*args, **kwargs)
-        if res.output and not res.output.endswith("/"):
+        if not res.output:
+            self.error("no --output location specified, aborting")
+        if not res.output.endswith("/"):
             res.output += "/"
         if res.url_prefix and not res.url_prefix.endswith("/"):
             res.url_prefix += "/"
@@ -546,10 +632,20 @@ class Sigal(Processor):
 
         for path in paths:
             logging.info("copying %s into %s", path, pictures_dir)
-            if not self.dryrun:
+            if self.dryrun:
+                continue
+            try:
                 shutil.copytree(
-                    path, str(pictures_dir), symlinks=not self.args.follow_symlinks
+                    path,
+                    str(pictures_dir),
+                    symlinks=not self.args.follow_symlinks,
+                    ignore=shutil.ignore_patterns(".*"),
                 )
+            except NotADirectoryError:
+                logging.error(
+                    "--gallery only works on directories, %s is not a directory", path
+                )
+                return []
 
         build_dir = output_dir / "_build"
         command = (
@@ -713,13 +809,72 @@ class ScreenshotProcessor(Processor):
         return []
 
 
+class Shotman(ScreenshotProcessor):
+    """The shotman processor will take a screenshot and return the file."""
+
+    def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
+        snap_path, target = self._select_paths(tmp_dir)
+        command = ["shotman", "--capture"]
+        # XXX: shotman doesn't support passing a specific area, for
+        # which we could then use slurp to select, then wait, then
+        # pass the region to shotman. since we can't do that, activate
+        # the timer only when we capture a whole output, since
+        # otherwise we have this weird effect of waiting for a
+        # timeout, then dropping to a selection
+        if self.select:
+            command.append("region")
+        else:
+            command.append("output")
+            self.gtk_wait_for_timeout()
+        if self.dryrun:
+            logging.warning("running in dry run, not taking screenshot")
+            return []
+        logging.info("taking screenshot with '%s'", shlex_join(command))
+        env = os.environ | {"XDG_SCREENSHOTS_DIR": str(snap_path.parent)}
+        success = subprocess.call(command, env=env) == 0
+        if not success:
+            notify_user("screenshot command failed, aborting")
+            return []
+        # find latest file in parent
+        #
+        # ideally, shotman would support a --file argument or show
+        # where it's saving files, but alas it isn't:
+        # https://todo.sr.ht/~whynothugo/shotman/35
+        # https://todo.sr.ht/~whynothugo/shotman/19
+        newest_path = None
+        for p in snap_path.parent.iterdir():
+            if newest_path is None:
+                newest_path = p
+                continue
+            if p.stat().st_mtime > newest_path.stat().st_mtime:
+                newest_path = p
+                continue
+        if newest_path is not None:
+            snap_path = newest_path
+        if not snap_path.exists():
+            notify_user("snapshot file missing, is maim installed?")
+            return []
+        if snap_path.stat().st_size <= 0:
+            notify_user("snapshot file is empty, aborting")
+            snap_path.unlink()
+            return []
+        return self.confirm_upload(snap_path, target)
+
+
 class Grim(ScreenshotProcessor):
     """The Grim processor will take a screenshot and return the file after
     copying it into a temporary directory.
 
     An alternative to this is shotman, not in Debian:
 
-    https://git.sr.ht/~whynothugo/shotman"""
+    https://git.sr.ht/~whynothugo/shotman
+
+    Another is grimshot, a shell script with interesting ideas like
+    pulling geometry from sway:
+
+    https://github.com/OctopusET/sway-contrib/blob/master/grimshot
+
+    """
 
     def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
         """wrap grim around a timer, preview and prompt"""
@@ -745,7 +900,7 @@ class Grim(ScreenshotProcessor):
             logging.warning("running in dry run, not taking screenshot")
             return []
 
-        logging.debug("running screenshot with '%s'", shlex_join(command))
+        logging.info("taking screenshot with '%s'", shlex_join(command))
         success = subprocess.call(command) == 0
         if not success:
             notify_user("screenshot command failed, aborting")
@@ -793,7 +948,7 @@ class Maim(ScreenshotProcessor):
             select_pipe.stdout.strip().decode("ascii"),
             str(snap_path),
         ]
-        logging.debug("running screenshot with %s", shlex_join(command))
+        logging.info("taking screenshot with %s", shlex_join(command))
         if not self.dryrun:
             success = subprocess.call(command) == 0
             if not success:
@@ -869,7 +1024,7 @@ class RsyncUploader(Uploader):
         # XXX: begin nasty rsync commandline logic
 
         # rsync is weird. it behaves differently when its arguments
-        # have trailling slashes or not. this specifically affects
+        # have trailing slashes or not. this specifically affects
         # directory (but technically, if you pass a file with a
         # trailing slash, it will fail, obviously)
         #
@@ -888,7 +1043,7 @@ class RsyncUploader(Uploader):
         # rsync -av /src/foo /dest
         # rsync -av /src/foo/ /dest/foo
 
-        # They ommitted, obviously, that this is also identical:
+        # They omitted, obviously, that this is also identical:
         #
         # rsync -av /src/foo/ /dest/foo/
         #
@@ -1020,6 +1175,7 @@ class History:
         """
         # date, time, token, uri
         entry = None
+        line = None
         with open(self.path, "r") as fp:
             for line in fp:
                 pass
@@ -1344,8 +1500,12 @@ def notify_user(message: str) -> None:
     command = ("notify-send", message)
     # do not fail on notifications
     logging.debug("running %s", command)
-    if subprocess.call(command) != 0:
-        logging.warning("failed to run command %r", command)
+    try:
+        subprocess.run(command, check=True)
+    except (subprocess.CalledProcessError, OSError) as e:
+        logging.warning(
+            "failed to run command %s, ignoring: %s", shlex_join(command), e
+        )
 
 
 def main() -> None:
@@ -1384,7 +1544,7 @@ def main() -> None:
 
     # 2. prepare the token
 
-    # ... from history (if --republish, which is implicity with --undo)
+    # ... from history (if --republish, which is implicit with --undo)
     if args.republish or args.command == "undo":
         last = history.last()
         # do not override provided token
@@ -1581,4 +1741,7 @@ def main() -> None:
 
 
 if __name__ == "__main__":
-    main()
+    try:
+        main()
+    except KeyboardInterrupt:
+        logging.warning("interrupted, aborting")
